/*
 * ext_io.c - Communication infrastructure for the external interface
 *
 * Written 2001 by Werner Almesberger
 * Copyright 2001 Network Robots
 */


#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <assert.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>

#include "config.h"
#include "util.h"
#include "error.h"
#include "ext.h"


static FILE *stream = NULL;	/* NULL if "ext" subsystem is not active */
static pid_t pid = 0;		/* 0 if there is no process */
static char in_path[100]; /* @@@ */
static char out_path[100]; /* @@@ */
static char **my_argv;		/* undefined if pid != 0 */

typedef struct _ext_arg {
    const char *name;
    const char *arg;
    struct _ext_arg *next;
} EXT_ARG;

static EXT_ARG *ext_args = NULL,**last_arg = &ext_args;


static void copy_argv(char **argv,const char *name)
{
    const EXT_ARG *ext_arg;
    int argc,i;

    for (argc = 0; argv[argc]; argc++);
    for (ext_arg = ext_args; ext_arg; ext_arg = ext_arg->next)
	if (!*ext_arg->name || !strcmp(ext_arg->name,name)) argc++;
    my_argv = malloc(sizeof(char *)*(argc+1));
    for (i = 0; argv[i]; i++)
	my_argv[i] = stralloc(argv[i]);
    for (ext_arg = ext_args; ext_arg; ext_arg = ext_arg->next)
	if (!*ext_arg->name || !strcmp(ext_arg->name,name))
	    my_argv[i++] = stralloc(ext_arg->arg);
    my_argv[i] = NULL;
}


static void free_argv(void)
{
    int i;

    for (i = 0; my_argv[i]; i++)
	free(my_argv[i]);
    free(my_argv);
}


void add_tcc_external_arg(const char *name,const char *arg)
{
    *last_arg = alloc_t(EXT_ARG);
    (*last_arg)->name = stralloc(name);
    (*last_arg)->arg = stralloc(arg);
    (*last_arg)->next = NULL;
    last_arg = &(*last_arg)->next;
}

static void child(char **argv)
{
    const char *path;
    int fd;

    path = getenv("PATH");
    if (path) {
	const char *tcng_topdir;
	char *new_path;

	tcng_topdir = getenv("TCNG_TOPDIR");
	if (!tcng_topdir) tcng_topdir = TOPDIR;
	new_path = alloc(strlen(path)+strlen(tcng_topdir)+strlen(TCNG_BIN_DIR)+
	  3);
	sprintf(new_path,"%s:%s/%s",path,tcng_topdir,TCNG_BIN_DIR);
	if (setenv("PATH",new_path,1) < 0) {
	    perror("setenv");
	    exit(1);
	}
    }
    (void) close(0);
    fd = open(in_path,O_RDONLY);
    if (fd < 0) {
	perror(in_path);
	exit(1);
    }
    if (dup2(fd,0) < 0) {
	perror("dup2");
	exit(1);
    }
    fd = creat(out_path,0644);
    if (fd < 0) {
	perror(out_path);
	exit(1);
    }
    if (dup2(fd,1) < 0) {
	perror("dup2");
	exit(1);
    }
    if (execvp(*argv,argv) < 0) {
	perror(*argv);
	exit(1);
    }
    /* not reached */
}


static void cleanup_external(void)
{
    if (stream) ext_close();
}


FILE *ext_open(char **argv,const char *name)
{
    static int need_exit_handler = 1;

    assert(!stream);
    sprintf(in_path,"__ext_%d.in",getpid());
    sprintf(out_path,"__ext_%d.out",getpid());
    stream = fopen(in_path,"w");
    if (!stream) {
	perror(in_path);
	exit(1);
    }
    if (need_exit_handler) {
	need_exit_handler = 0;
	atexit(cleanup_external);
    }
    copy_argv(argv,name);
    return stream;
}


int ext_read(void)
{
    int status;

    assert(stream);
    (void) fflush(stream);
    if (ferror(stream)) error("error writing to sub-process");
    if (!freopen(DEV_NULL,"r",stream)) { /* close old stream */
	perror("/dev/null");
	exit(1);
    }
    fflush(stdout);
    fflush(stderr);
    pid = fork();
    if (!pid) child(my_argv);
    if (waitpid(pid,&status,0) < 0) {
	perror("waitpid");
	exit(1);
    }
    if (!freopen(out_path,"r",stream)) {
	perror(out_path);
	exit(1);
    }
    return WIFEXITED(status) ? WEXITSTATUS(status) : -1;
}


void ext_close(void)
{
    assert(stream);
    if (fclose(stream) == EOF) {
	perror("fclose");
	exit(1);
    }
    stream = NULL;
    if (unlink(in_path) < 0) {
	perror(in_path);
	exit(1);
    }
    free_argv();
    if (pid) {
	if (unlink(out_path) < 0) {
	    perror(out_path);
	    exit(1);
	}
	pid = 0;
    }
}
